/*	$OpenBSD: malo.h,v 1.9 2006/11/24 20:45:33 mglocker Exp $ */

/*
 * Copyright (c) 2006 Claudio Jeker <claudio@openbsd.org>
 * Copyright (c) 2006 Marcus Glocker <mglocker@openbsd.org>
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

#define	VENDORID_MARVELL	0x11ab
#define	DEVICEID_MRVL_88W8310	0x1fa7
#define	DEVICEID_MRVL_88W8335R1	0x1faa
#define	DEVICEID_MRVL_88W8335R2	0x1fab

/* internal structures and defines */
struct malo_node {
	struct ieee80211_node		ni;
};

struct malo_rx_data {
	bus_dmamap_t	map;
	struct mbuf	*m;
};

struct malo_tx_data {
	bus_dmamap_t		map;
	struct mbuf		*m;
	uint32_t		softstat;
	struct ieee80211_node	*ni;
};

/* RX descriptor used by HW */
struct malo_rx_desc {
	uint8_t		rxctrl;
	uint8_t		rssi;
	uint8_t		status;
	uint8_t		channel;
	uint16_t	len;
	uint8_t		reserved1;	/* actually unused */
	uint8_t		datarate;
	uint32_t	physdata;	/* DMA address of data */
	uint32_t	physnext;	/* DMA address of next control block */
	uint16_t	qosctrl;
	uint16_t	reserved2;
} __packed;

/* TX descriptor used by HW */
struct malo_tx_desc {
	uint32_t	status;
	uint8_t		datarate;
	uint8_t		txpriority;
	uint16_t	qosctrl;
	uint32_t	physdata;	/* DMA address of data */
	uint16_t	len;
	uint8_t		destaddr[6];
	uint32_t	physnext;	/* DMA address of next control block */
	uint32_t	reserved1;	/* SAP packet info ??? */
	uint32_t	reserved2;
} __packed;

#define MALO_RX_RING_COUNT	256
#define MALO_TX_RING_COUNT	256
#define MALO_MAX_SCATTER	8	/* XXX unknown, wild guess */
#define	MALO_MAX_TX_SEGSIZE	(65535 + 200) /* XXX shoudl be changed */

#define	MALO_CMD_DMA_ALIGN	PAGE_SIZE	/* XXX guessed value */
#define	MALO_CMD_DMA_SZ		4096	/* XXX guessed value */
#define	MALO_RING_ALIGN		256	/* just guessing */
#define	MALO_TX_RING_SZ	(sizeof(struct malo_tx_desc) * MALO_TX_RING_COUNT)
#define	MALO_RX_RING_SZ	(sizeof(struct malo_rx_desc) * MALO_RX_RING_COUNT)
#define	MALO_TX_RING_INC(x)	((x) = ((x) + 1) % MALO_TX_RING_COUNT)
#define	MALO_RX_RING_INC(x)	((x) = ((x) + 1) % MALO_RX_RING_COUNT)

/*
 * Firmware commands
 */
#define MALO_CMD_GET_HW_SPEC		0x0003
#define MALO_CMD_SET_WEPKEY		0x0013
#define MALO_CMD_SET_RADIO		0x001c
#define MALO_CMD_SET_AID		0x010d
#define MALO_CMD_SET_TXPOWER		0x001e
#define MALO_CMD_SET_ANTENNA		0x0020
#define MALO_CMD_SET_PRESCAN		0x0107
#define MALO_CMD_SET_POSTSCAN		0x0108
#define MALO_CMD_SET_RATE		0x0110
#define MALO_CMD_SET_CHANNEL		0x010a
#define MALO_CMD_SET_RTS		0x0113
#define MALO_CMD_SET_SLOT		0x0114
#define MALO_CMD_RESPONSE		0x8000

#define MALO_CMD_RESULT_OK		0x0000	/* everything is fine */
#define MALO_CMD_RESULT_ERROR		0x0001	/* general error */
#define MALO_CMD_RESULT_NOSUPPORT	0x0002	/* command not valid */
#define MALO_CMD_RESULT_PENDING		0x0003	/* will be processed */
#define MALO_CMD_RESULT_BUSY		0x0004	/* command ignored */
#define MALO_CMD_RESULT_PARTIALDATA	0x0005	/* buffer too small */

struct malo_cmdheader {
	uint16_t	cmd;
	uint16_t	size;		/* size of the command, incl. header */
	uint16_t	seqnum;		/* seems not to matter that much */
	uint16_t	result;		/* set to 0 on request */
	/* following the data payload, up to 256 bytes */
};

struct malo_hw_spec {
	uint16_t	HwVersion;
	uint16_t	NumOfWCB;
	uint16_t	NumOfMCastAdr;
	uint8_t		PermanentAddress[6];
	uint16_t	RegionCode;
	uint16_t	NumberOfAntenna;
	uint32_t	FWReleaseNumber;
	uint32_t	WcbBase0;
	uint32_t	RxPdWrPtr;
	uint32_t	RxPdRdPtr;
	uint32_t	CookiePtr;
	uint32_t	WcbBase1;
	uint32_t	WcbBase2;
	uint32_t	WcbBase3;
} __packed;

struct malo_cmd_wepkey {
	uint16_t	action;
	uint8_t		len;
	uint8_t		flags;
	uint16_t	index;
	uint8_t		value[IEEE80211_KEYBUF_SIZE];
	uint8_t		txmickey[IEEE80211_WEP_MICLEN];
	uint8_t		rxmickey[IEEE80211_WEP_MICLEN];
	uint64_t	rxseqctr;
	uint64_t	txseqctr;
} __packed;

struct malo_cmd_radio {
	uint16_t	action;
	uint16_t	preamble_mode;
	uint16_t	enable;
} __packed;

struct malo_cmd_aid {
	uint16_t	associd;
	uint8_t		macaddr[6];
	uint32_t	gprotection;
	uint8_t		aprates[14];
} __packed;

struct malo_cmd_txpower {
	uint16_t	action;
	uint16_t	supportpowerlvl;
	uint16_t	currentpowerlvl;
	uint16_t	reserved;
	uint16_t	powerlvllist[8];
} __packed;

struct malo_cmd_antenna {
	uint16_t	action;
	uint16_t	mode;
} __packed;

struct malo_cmd_postscan {
	uint32_t	isibss;
	uint8_t		bssid[6];
} __packed;

struct malo_cmd_channel {
	uint16_t	action;
	uint8_t		channel;
} __packed;

struct malo_cmd_rate {
	uint8_t		dataratetype;
	uint8_t		rateindex;
	uint8_t		aprates[14];
} __packed;

struct malo_cmd_slot {
	uint16_t	action;
	uint8_t		slot;
} __packed;

#define malo_mem_write4(_sc, off, x) \
	bus_write_4((_sc)->sc_res_mem[0], (off), (x))
#define malo_mem_write2(_sc, off, x) \
	bus_write_2((_sc)->sc_res_mem[0], (off), (x))
#define malo_mem_write1(_sc, off, x) \
	bus_write_1((_sc)->sc_res_mem[0], (off), (x))

#define malo_mem_read4(_sc, off) \
	bus_read_4((_sc)->sc_res_mem[0], (off))
#define malo_mem_read1(_sc, off) \
	bus_read_1((_sc)->sc_res_mem[0], (off))

#define malo_ctl_write4(_sc, off, x) \
	bus_write_4((_sc)->sc_res_mem[1], (off), (x))
#define malo_ctl_read4(_sc, off) \
	bus_read_4((_sc)->sc_res_mem[1], (off))
#define malo_ctl_read1(_sc, off) \
	bus_read_1((_sc)->sc_res_mem[1], (off))

#define malo_ctl_barrier(_sc, t) \
	bus_barrier((_sc)->sc_res_mem[1], 0x0c00, 0xff, (t))

struct malo_rx_ring {
	bus_dma_tag_t		ring_tag;
	bus_dmamap_t		ring_map;
	bus_addr_t		physaddr;
	bus_dma_tag_t		rx_tag;
	bus_dmamap_t		rx_map;
	bus_dmamap_t		rx_sparemap;
	struct malo_rx_desc	*desc;
	struct malo_rx_data	data[MALO_RX_RING_COUNT];
	/* int			count XXX not used in FreeBSD; */
	int			cur;
	int			next;
};

struct malo_tx_ring {
	bus_dma_tag_t		ring_tag;
	bus_dmamap_t		ring_map;
	bus_addr_t		physaddr;
	bus_dma_tag_t		tx_tag;
	bus_dmamap_t		tx_map;
	struct malo_tx_desc	*desc;
	struct malo_tx_data	data[MALO_TX_RING_COUNT];
	/* int			count; XXX not used in FreeBSD */
	int			queued;
	int			cur;
	int			next;
	int			stat;
};

#define	MALO_TX_RING_ADDR(sc, i)	\
	((sc)->sc_txring.physaddr + sizeof(struct malo_tx_desc) * (i))
#define	MALO_RX_RING_ADDR(sc, i)	\
	((sc)->sc_rxring.physaddr + sizeof(struct malo_rx_desc) * (i))

#define MALO_RX_RADIOTAP_PRESENT					\
	((1 << IEEE80211_RADIOTAP_FLAGS) |				\
	 (1 << IEEE80211_RADIOTAP_CHANNEL) |				\
	 (1 << IEEE80211_RADIOTAP_DBM_ANTSIGNAL))

struct malo_rx_radiotap_hdr {
	struct ieee80211_radiotap_header	wr_ihdr;
	uint8_t					wr_flags;
	uint16_t				wr_chan_freq;
	uint16_t				wr_chan_flags;
	uint8_t					wr_rssi;
#if 0
	/* XXX */
	uint8_t					wr_max_rssi;
#endif
} __packed;

#define MALO_TX_RADIOTAP_PRESENT					\
	((1 << IEEE80211_RADIOTAP_FLAGS) |				\
	 (1 << IEEE80211_RADIOTAP_RATE) |				\
	 (1 << IEEE80211_RADIOTAP_CHANNEL))

struct malo_tx_radiotap_hdr {
	struct ieee80211_radiotap_header	wt_ihdr;
	uint8_t					wt_flags;
	uint8_t					wt_rate;
	uint16_t				wt_chan_freq;
	uint16_t				wt_chan_flags;
} __packed;

#define MALO_RESOURCE_MAX	2
#define MALO_MSI_MESSAGES	1

struct malo_softc {
	struct ifnet		*sc_ifp;
	device_t		sc_dev;
	struct resource_spec 	*sc_mem_spec;
	struct resource		*sc_res_mem[MALO_RESOURCE_MAX];
	struct resource_spec	*sc_irq_spec;
	struct resource		*sc_res_irq[MALO_MSI_MESSAGES];
	void			*sc_intrhand[1];
	struct taskqueue	*sc_tq;
	struct task		sc_int_task;
	int			sc_msi;
	struct ieee80211com	sc_ic;

	struct malo_rx_ring	sc_rxring;
	struct malo_tx_ring	sc_txring;

	bus_dma_tag_t		sc_parent_tag;

	bus_dma_tag_t		sc_cmd_tag;
	bus_dmamap_t		sc_cmd_map;
	void			*sc_cmd_mem;
	bus_addr_t		sc_cmd_dmaaddr;
	uint32_t		*sc_cookie;
	bus_addr_t		sc_cookie_dmaaddr;

	uint32_t		sc_RxPdWrPtr;
	uint32_t		sc_RxPdRdPtr;

	int			(*sc_newstate)
				(struct ieee80211com *,
				 enum ieee80211_state, int);

	struct callout		sc_scan_to;
	int			sc_tx_timer;
	int			sc_last_txrate;

	struct bpf_if		*sc_drvbpf;

	union {
		struct malo_rx_radiotap_hdr th;
		uint8_t pad[64];
	}		sc_rxtapu;
#define sc_rxtap	sc_rxtapu.th
	int		sc_rxtap_len;

	union {
		struct malo_tx_radiotap_hdr th;
		uint8_t pad[64];
	}		sc_txtapu;
#define sc_txtap	sc_txtapu.th
	int		sc_txtap_len;
	struct mtx	sc_mtx;
};

#define MALO_LOCK(_sc)		mtx_lock(&(_sc)->sc_mtx)
#define MALO_UNLOCK(_sc)	mtx_unlock(&(_sc)->sc_mtx)
#define MALO_LOCK_ASSERT(_sc)	mtx_assert(&(_sc)->sc_mtx, MA_OWNED)
#define MALO_UNLOCK_ASSERT(_sc)	mtx_assert(&(_sc)->sc_mtx, MA_NOTOWNED)
